#![deny(
    missing_docs,
    missing_debug_implementations,
    missing_copy_implementations,
    trivial_casts,
    trivial_numeric_casts,
    unsafe_code,
    unstable_features,
    unused_import_braces,
    unused_qualifications
)]

//!
//! libFenrir is the official rust library implementing the Fenrir protocol

use ::std::{
    io::Result,
    net::{IpAddr, Ipv4Addr, SocketAddr},
};
use ::tokio::{net::UdpSocket, runtime};
use ::tracing::error;

mod config;
pub use config::Config;

/// Instance of a fenrir endpoint
#[allow(missing_copy_implementations, missing_debug_implementations)]
pub struct Fenrir {
    /// library Configuration
    cfg: Config,
    /// internal runtime
    rt: ::tokio::runtime::Runtime,
}

/// Initialize a new Fenrir endpoint
pub fn init(config: &Config) -> Result<Fenrir> {
    // PERF: linux iouring
    // PERF: multithread pinning with hwloc2 for topology
    let f = Fenrir::new(config.clone())?;
    f.add_listener();
    Ok(f)
}

impl Fenrir {
    /// Create a new Fenrir endpoint
    fn new(config: Config) -> Result<Self> {
        let mut builder = runtime::Builder::new_multi_thread();
        if let Some(threads) = config.threads {
            builder.worker_threads(threads.get());
        }
        let rt = match builder.thread_name("libFenrir").build() {
            Ok(rt) => rt,
            Err(e) => {
                ::tracing::error!("Can't initialize: {}", e);
                return Err(e);
            }
        };
        Ok(Fenrir {
            cfg: config,
            rt: rt,
        })
    }
    /// Add an UDP listener
    fn add_listener(&self) {
        if self.cfg.listen.len() == 0 {
            self.rt.spawn(listen_udp(SocketAddr::new(
                IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0)),
                0, // 0 means random port
            )));
        } else {
            for sock in self.cfg.listen.iter() {
                self.rt.spawn(listen_udp(sock.clone()));
            }
        }
    }
}
/// Add an async udp listener
async fn listen_udp(sock: SocketAddr) -> Result<()> {
    UdpSocket::bind(sock).await?;
    Ok(())
}
